
/********************************************************************************
 *										*
 *				 M A C E K					*
 *		("Matroids Also Computed Efficiently" Kit)			*
 *										*
 * A set of tools and routines for computations with representable matroids.	*
 * Copyright (C) 2001,2002  Petr Hlineny.					*
 *										*
 *   This program is free software; you can redistribute it and/or modify	*
 *   it under the terms of the GNU General Public License as published by	*
 *   the Free Software Foundation; either version 2 of the License, or		*
 *   (at your option) any later version.					*
 *   You should have received a copy of the GNU General Public License		*
 *   along with this program; if not, write to the Free Software		*
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.	*
 * 										*
 * See the top-level "ABOUT" and "doc/ *" files for more information.		*
 *										*
 ********************************************************************************/














/**
 * Some theory about generating matroids for start...
 * The master description of the generating process is in ../include/gener.h.
 * 
 * This file contains the main generating functions.
 * The function gener_extframe_ext() is the general interface for this process.
 * (It converts relevant information from the given frame to the elim-sequence,
 * and calls the internal extension-functions.
 * The resulting elim-sequences are then stored back into a list of frames.)
 * 
 * The other functions here control the generating process,
 * according to the elimination tests (see also genstep.c).
 * 
 * .......... more about ctrl-modifications to generating process .........
 * 
 * 
**/


#include "macek.h"
#include "gen.h"















/******************	Extensions of a frame	*******************/
/******************************************************************/


//#undef CURDLEV
#define	CURDLEV		5		/* (use <=DEBUGLEV+2 to perform extra checks) */



/**
 * This function extends the matrix of the given frame fr by a column if trq==1,
 * or by a row if trq==0.
 * The optional parameter ctrl may be used to control the generating process
 * (default 0, see gener_extensions() for more).
 * The return value if a list of new frames holding the generated matrices,
 * and having no parents or sons.
 * 
 * The parameteres of the generating process are given by options in the frame fr.
 * The two main options (that must be present right in the frame fr, no inheritance)
 * are GEN_OPTPREFIX"bsize" and GEN_OPTPREFIX"signature".
 * They determine the base minor dimensions (top-left submatrix of fr) and the sequence
 * signature, as defined in ../include/gener.h.
 * These options are also automatically stored to the new frames (signature updated).
 * The new frames also get names determined by the pattern GEN_??EXTNAME, and comments.
 * The sequence signature should never be set by hand(!), but taken from previous computations.
 * 
 * More attributes of the elimination sequence are obtained next from other options.
 * (These options are already inherited, and they are not automatically written to the
 * new frames.)
 * For example, one may give list of forbidden minors, length of forbidden fan, etc.
 * See the description of options in ../frame/fropts.c.
 * It is possible to implement more options for an elimination sequence in the function
 * gener_extframe_incl() in gener-more.inc.
 * To write requested options to the generated new frames, one must list the option
 * names within the options "extinherit" and "extinheritall".
 * 
 * More information in ../include/gener.h...
**/

framestruc**	gener_extframe_ext(framestruc *fr, int trq, int ctrl) {
	ematrix		*e, *emin, **eel, **eml;
	elimseqm	*q, **qel,**qq;
	framestruc	*frc,*frn, **flist;
	optionstruc	*op, **oih;
	int		i,j,b, bsz[3];
	char		**ml, **s, buf[400],buf2[30];
	
	if (FRMATRIX(fr)==NULL)  {PROGERROR("No matrix given to gener_extframe()");  return NULL;}
	e = ematrix_copy_notr(FRMATRIX(fr));
	ematrix_resetid(e);	/* (must repair the id labels after clearing the transpose state) */
	DEBUG(CURDLEV-2,"Calling to get %sextensions of the frame %p [%s], (ct %d)...\n",trq?"column ":"row co-",fr,FRNAME(fr),ctrl);
	EMATDEBUG(CURDLEV,e,"\t\t");
	
		/**
		 * Here we set up the basic attributes of a new elimination sequence
		 * according to the current frame fr.
		 * We find out the base-minor size in the frame, and the sequence signature.
		 * We also set the name for the sequence by fr, but this name is not copied(!).
		 * These attributes are kept together by the elim structure q.
		 * (The matrix e and the base minor emin are freed from q at the end.)
		**/
	q = new_elimseqm();
	i = frame_getoptionnum_noanc(fr,GEN_OPTPREFIX"bsize",bsz,2,0);
	if (!i) {
		DEBUG(CURDLEV-2,"## no base minor is given, assuming equal to the whole matrix...\n");
		emin = e;
	} else {
		if (bsz[0]<=0)  bsz[0] += ROWSM(e);  if (bsz[1]<=0)  bsz[1] += COLSM(e);
		DEBUG(CURDLEV-1,"- base minor of size %d x %d is given...\n",bsz[0],bsz[1]);
		emin = ematrix_refer(e,0,bsz[0],0,bsz[1]);
	}
	i = frame_getoptionnum_noanc(fr,GEN_OPTPREFIX"signature",bsz,1,0);
	if (!i) {
		bsz[0] = (1<<(COLSM(e)-COLSM(emin)))-1;
		DEBUG(CURDLEV-2,"## no elim-sequence signature is given, assuming %d...\n",bsz[0]);
	}
	//********* possibly test the signature first... - but the signature is never set by hand!
	q = elimseqm_setfor_min(q,e,emin,bsz[0]);
	if (emin && emin!=e)  dispose_ematrix(emin);
	ESNAME(q) = FRNAME(fr);		/* (do not ever free the name from q !!!) */
	
		/**
		 * Additional attributes are taken from fr here as matrix-lists.
		 * That is a list of forbidden minors, and a list of minors for a "tight major".
		 * These lists are not copied or modified in the generating process.
		 * (The lists are disposed at the end here.)
		**/
	ml = frame_getoptionval_all(fr,GEN_OPTPREFIX"forbid");
	for (eml=NULL, s=ml; s?*s:0; s++) {
		eel = frame_inputmatrices(*s);
		if (eel)  eml = alist_applist(eml,eel);
	}
	if (ml)  dispose_alist(ml);
	if (eml)  DEBUG(CURDLEV-1,"- list of forbidden minors of length %d is given... [%s %s ]\n",
			alist_getlength(eml),eml[0]?EMNAME(eml[0]):"",eml[0]&&eml[1]?EMNAME(eml[1]):"");
	if (eml) if (struct_hasminorlist(e,eml)) {
		dispose_alist_mats(eml);  eml = NULL;
		USERERROR("The matrix for extending already has a forbidden minor in the list! Clearing.");
	}
	if (eml)  ESFORBID(q) = eml;
	ml = frame_getoptionval_all(fr,GEN_OPTPREFIX"tight");
	for (eml=NULL, s=ml; s?*s:0; s++) {
		eel = frame_inputmatrices(*s);
		if (eel)  eml = alist_applist(eml,eel);
	}
	if (ml)  dispose_alist(ml);
	if (eml)  DEBUG(CURDLEV-1,"- list of \"tight-major\" minors of length %d is given... [%s %s ]\n",
			alist_getlength(eml),eml[0]?EMNAME(eml[0]):"",eml[0]&&eml[1]?EMNAME(eml[1]):"");
	if (eml) if (struct_hasdelcontr(e,eml)) {
		dispose_alist_mats(eml);  eml = NULL;
		USERERROR("The matrix for extending is not a tight major of the list! Clearing.");
	}
	if (eml)  ESTIGHT(q) = eml;
	
		/**
		 * More attributes (number-valued) are obtained here... (like forbidden fans)
		 * 
		 * Other parameters........
		**/
	i = frame_getoptionnum(fr,GEN_OPTPREFIX"nofan",bsz,1,0);
	if (i && bsz[0]<1111 && bsz[0]>=3)  ESNOFAN(q) = bsz[0];
	if (i)  DEBUG(CURDLEV-1,"- no fan of length >=%d in the elim sequence, given...\n",ESNOFAN(q));
	
	//********** tails, and more.........
	
		/**
		 * This part calls a third-party modification function gener_extframe_incl(),
		 * for more attributes to the sequence q.
		 * Then the generating function gener_extensions() is called, storing the
		 * generated list of (next) sequences in qel.
		 * The next call to third-party modification function gener_extframe_incl()
		 * then may prepare a "master copy" of the output frame frc.
		**/
	q = gener_extframe_incl(fr,trq,ctrl,q,NULL,NULL,NULL,1);
	qel = (elimseqm**)gener_extensions(1,ctrl,q,trq);
	frc = new_frame(NULL);  flist = NULL;
	q = gener_extframe_incl(fr,trq,ctrl,q,qel,&frc,NULL,0);
	FRPFINDEX(frc) = FRPFINDEX(fr);
	
		/**
		 * This part inherits the options requested by "extinherit??"
		 * for the master frame frc.
		**/
	ml = frame_getoptionval_all(fr,"extinherit");
	oih = frame_getoptionlist_last(fr,ml);
	if (ml)  dispose_alist(ml);
	FROPTIONS(frc) = alist_applist(FROPTIONS(frc),oih);
	ml = frame_getoptionval_all(fr,"extinheritall");
	oih = frame_getoptionlist_all(fr,ml);
	if (ml)  dispose_alist(ml);
	FROPTIONS(frc) = alist_applist(FROPTIONS(frc),oih);
		/**
		 * This part converts the generated list of sequences to output frames.
		 * We set the "signature" option accordingly for the new frames, also
		 * give the new names by GEN_??EXTNAME and set a simple comment.
		 * The options previously set to the master frame frc are also copied.
		 * (Matrices of the elim-sequences are given to the frames, and the
		 * elim structures themselves are later freed with their list qel.)
		**/
	for (qq=qel,i=1; qq?*qq:0; qq++,i++) {
		frn = frame_copy_full(frc);
		frame_setmatrix(frn,ESMATRIX(*qq));	/* (the matrix is not copied there) */
		op = new_option_onenum(GEN_OPTPREFIX"signature",ESELIMSG(*qq));
		frame_addoption(frn,op,2);
		op = new_option_twonum(GEN_OPTPREFIX"bsize",ROWSM(ESMINOR(*qq)),COLSM(ESMINOR(*qq)));
		frame_addoption(frn,op,3);
		snprintf(buf,380,trq?GEN_EXTNAME:GEN_COEXTNAME,FRNAME(fr),i);
		buf[380] = 0;  frame_setname(frn,buf);
		for (j=ESLENGTH(*qq)-1, b=ESELIMSG(*qq); j>=0; j--, b/=2)
			buf2[j] = (b%2)?'c':'r';
		buf2[ESLENGTH(*qq)] = 0;
		snprintf(buf,380,"mat #%d %sexten to \'%.12s\', base %dx%d, sig %s(%ld)",
			i,trq?"column ":"row co-",FRNAME(fr),ROWSM(ESMINOR(*qq)),COLSM(ESMINOR(*qq)),buf2,(long)ESELIMSG(*qq));
		buf[380] = 0;  FRCOMMENT(frn) = MSTRDUP(buf);
		flist = alist_append(flist,frn);
	}
	if (FRPARENT(frc) || FRSONS(frc))  {PROGERROR("the master frame frc should not have parent or sons %p %p %p",frc,FRPARENT(frc),FRSONS(frc));}
	dispose_frame(frc);
	
		/**
		 * Here the third-party modification is called again to (possibly) add
		 * more options to the output frames in flist, and to destroy its local data.
		 * Finally, the data associated with the elimination sequence are freed.
		 * (Recall that this data is not copied, just refered in the ext sequences.)
		 * The matrices from qel were already stored in the output frames.
		**/
	q = gener_extframe_incl(fr,trq,ctrl,q,qel,NULL,flist,-1);
	if (ESFORBID(q))  dispose_alist_mats(ESFORBID(q));
	if (ESTIGHT(q))  dispose_alist_mats(ESTIGHT(q));
	dispose_ematrix(ESMINOR(q));  dispose_ematrix(ESMATRIX(q));
	FREE(q);
	dispose_alist(qel);	/* the elim-sequences are freed with their list(!) */
	return flist;
}













/******************	Generating matrix extensions	*******************/
/**************************************************************************/


#undef CURDLEV
#define	CURDLEV		5	/* (>=5 is OK for randomized checks, use <=DEBUGLEV+2 for extra checks) */


		/* these variables are used for profiling output of the generating process: */
int	stseq3, ststruc3, stcanon3, ststruc2, ststruc1, stcanon1, stseq0, ststruc0, stcanon0;


/**
 * This function generates (co)extensions to the matrix in the given elim sequence q.
 * All distinct extension lines are generated by calling gener_matextens_..().
 * The generated extensions are then tested by the genstep_????check() functions from
 * genstep.c (see gener_extensions_pass()), and only proper extensions pass through.
 * Internal parameters of the sequence structure q are in control of these tests.
 * The function works only for connected(!) matrices if in a partial field.
 * 
 * If trq==1, then we compute column extensions, while for trq==0 row co-extensions
 * (both against the 0-trasnposition state of the matrix).
 * We use "column implementation", so the matrices are first transposed for row coextension.
 * The parameter ctrl controls what tests are performed:
 * ctrl==0 means all tests, ctrl>0 only some (or no) canonical tests or no pre-tests.
 * (Use macros like GCTRL_NOPRECHECK, GCTRL_CANONLEV(l) in "or" expr.)
 * Read more about performing canonical tests in gener.h ...
 * If ch>0+, then some matrices are printed along.
 * 
 * The function returns a list (alist) of pointers to new sequence structures holding
 * all generated and passed matrices, with internal parameters copied from the sequence q.
 * (Remember to free the list and the structures later; but not the refered data.)
**/


void**	gener_extensions(int ch, int ctrl, void *vq, int trq) {
	int		i,k,r, npas=0, tri;
	ematrix		*ei,*ee, **exl=NULL, **x=NULL;
	elimseqm	*q=vq, *qn, *qo;
	void		**out;
	
	ei = ESMATRIX(q);  tri = ISTRANSPM(ei);
#ifndef FASTPROG
	if (!ei || (ESMINOR(q)? (ROWSM(ESMINOR(q))<=0||COLSM(ESMINOR(q))<=0):1))
		{PROGERROR("Call only for nonempty matrix %p and nonempty base minor %p !",ei,ESMINOR(q)); return NULL;}
	if (ISTRANSPM(ei))  {PROGERROR("Do not start an elim sequence with a trasposed matrix %p",ei); return NULL;}
	if (ch<-1)  {PROGERROR("Do not call with ch==%d <-1",ch);  ch = -1;}
	DEBUG(CURDLEV-1,"Generating %sextensions of a %dx%d matrix %p [%s] ...\n",trq?"column ":"row co-",ROWSM(ei),COLSM(ei),ei,ESNAME(q));
	DEBUG(CURDLEV-1,"   - canonlev=%d, nopretest %c\n",GCTRL_CANONLEV(ctrl),(GCTRL_NOPRETEST&ctrl)?'Y':'N');
	test_elimseqm(q);
	if (ch==1)  ELIMQDEBUGS(CURDLEV-1,q," ==\t");  else if (ch>1)  ELIMQDEBUG(CURDLEV-1,q," ==\t");
	stseq3 = ststruc3 = stcanon3 = ststruc1 = ststruc2 = stcanon1= stseq0 = ststruc0 = stcanon0 = 0;
#endif		/* (the static variables stseq?... are used for debug statistics of the computation) */
	
			/**
			 * We prepare the variables for generating extensions:
			 * ei is the given matrix in q, qn is the new elim sequence which is
			 * later copied for the output, and exl gets the list of all possible
			 * matrix extensions generated in gener_matextens().
			**/
	ematrix_transpose_set(ei,!trq);		/* transposing for row co-extensions */
	out = new_alist(32);
	ee = ematrix_append_col(ematrix_copy(ei));
	qn = elimseqm_next(q,ee,trq);		/* the next elim seq structure - to be copied for output */
	if (genstep_seqsigncheck(qn,ESELIMSG(qn))<0) {
		DEBUG(CURDLEV-2,"## NOT accepting the next sequence signature %lX !\n",ESELIMSG(qn));
		exl = NULL;  k = -1;
	} else  k = 0;
	if (k>=0) {				/* getting all matrix (column-) extensions to ei */
		x = exl = gener_matextens_ch(ch,ei);
		if (exl?!*exl:1)  k = -1;
	}
			/**
			 * We cycle through all generated matrix extensions, and look whether
			 * they pass the tests in gener_extensions_pass().
			 * If so, then a new copy of the elim sequence qn is created for the
			 * matrix, and stored in the output list (in the original tr-state).
			 * But first genstep_pretest_start() is called to prepare some data
			 * used in faster generating checks.
			 * Finally, some quick tests of the sequence are run.
			**/
	if (k>=0 && !(ctrl&GCTRL_NOPRETEST))
		genstep_pretest_start(qn,1);	/* (now we use transpose-value 1 - a column) */
	
	if (k>=0)  for (x=exl; *x && k>=0; x++) {
#ifndef FASTPROG
		if (CURDLEV-0<=DEBUGLEV && ch>=0) {
			DEBUG(CURDLEV-0,"> extension #(%d) of [%s] trying: ",(int)(x-exl),ESNAME(q));
			for (i=0; i<ROWSM(ee); i++)  SDEBUG(CURDLEV-0," %s,",pfield_pvalue(6,EXPM(*x,i,COLSM(*x)-1),SIGNM(*x,i,COLSM(*x)-1)));
			SDEBUG(CURDLEV-0,"\n"); }
#endif
		r = gener_extensions_pass(ch,ctrl,qn,1,*x);
		if (r<0)  continue;
		++npas;				/* a passed matrix...: */
		qo = elimseqm_copyfor(qn,ematrix_copy(*x));
		ematrix_transpose_set(ESMATRIX(qo),tri);
		out = alist_append(out,qo);
#ifndef FASTPROG
		if (CURDLEV-1<=DEBUGLEV && ch>=0) {
			DEBUG(CURDLEV-1,"=>> extension #%d (%d) of [%s] accepted: ",npas,(int)(x-exl),ESNAME(q));
			for (i=0; i<ROWSM(ee); i++)  SDEBUG(CURDLEV-1," %s,",pfield_pvalue(6,EXPM(*x,i,COLSM(*x)-1),SIGNM(*x,i,COLSM(*x)-1)));
			SDEBUG(CURDLEV-1,"\n"); }
		if (ch>0)  EMATDEBUG(CURDLEV-1,ESMATRIX(qo),"\t++\t");
				/* some final debug testing (even in the fastprog mode) */
		if (IFRANDDEBUGLESS(60))  test_elimseqm(qo);
		else if (DEBUGLEV>2 || IFRANDDEBUG(8))  test_elimseqm_rand(qo);  else
#endif
		test_elimseqm_notough(qo);	/* (always at least the fast test...) */
	}
	
	if (k>=0 && !(ctrl&GCTRL_NOPRETEST))	/* (freeing precomputed information for passing tests) */
		genstep_pretest_stop(qn,1);	
	FREE(qn);  dispose_ematrix(ee);		/* (freeing what was used in generating extensions) */
	if (exl)  dispose_alist_mats(exl);
	ematrix_transpose_set(ei,tri);
	
	DEBUG(CURDLEV-2,"- Done, passed %d out of %d (co)extensions of the %dx%d matrix %p [%s]\n",npas,(int)(x-exl),ROWSM(ei),COLSM(ei),ei,ESNAME(q));
	SDEBUG(CURDLEV-2,"\t\t(seq3=%d, struc3=%d, canon3=%d, struc2=%d, struc1=%d, canon1=%d, seq0=%d, struc0=%d, canon0=%d)\n",stseq3,ststruc3,stcanon3,ststruc2,ststruc1,stcanon1,stseq0,ststruc0,stcanon0);
	SDEBUG(CURDLEV-1,"\t====================================================================================================\n");
	SDEBUG(CURDLEV-1,"\n");
	return out;		/* the output array of sequences is returned, terminated by NULL */
}



/**
 * Here we get a generated extension ee of the sequence qn, and we test whether this
 * extension passes all requirements of the sequence qn.
 * Remember that the last added line is a column of ee, and so tr==1.
 * The parameter ctrl controls what tests are performed:
 * ctrl==0 mean all tests, ctrl>0 only some (or no) canonical tests, or no pre-tests.
 * Read more about performing canonical tests in gener.h ...
 * 
 * The issued value of lev determines the level of tests - 0(implicit) means to do all tests thoroughly,
 * while values >0 (1,2,3 for now) request faster but incomplete tests - to speed up rejection.
 * The function returns -1 for a rejected matrix, and 1 for an accepted one.
**/

int	gener_extensions_pass(int ch, int ctrl, elimseqm *qn, int tr, ematrix *ee) {
	int		i, r=0;
	ematrix		*eqo;
	
	junk = ch;		/* (not really useful here) */
	if (!ee || tr!=1)  {PROGERROR("Wrong parameters %p %d",ee,tr);}
	eqo = ESMATRIX(qn);
	qn = elimseqm_setfor_mat(qn,ee);
	
		/* setting the extended line into canonical scale (unit-leading form): */
	if (r>=0)  r = genstep_linecheck_set(qn,ee,1);
	
		/**
		 * These are all partial-tests for faster rejection (at levels >0):
		 * If these tests reject a matrix, then it is for sure, but some bad matrices
		 * may still pass through.
		 * The line-check has been alredy performed above together with unit-scaling.
		 * The full tests are carried out in the next section.
		**/
	if (r>=0)  r = genstep_sequencecheck_lev(3,qn,ee,1);
	if (r>=0)  stseq3++;
	if (r>=0)  r = genstep_structurecheck_lev(3,qn,ee,1);
	if (r>=0)  ststruc3++;
	if (r>=0 && GCTRL_CANONLEV(ctrl)<=3)  r = genstep_canonicalcheck_lev(3,qn,ee,1);
	if (r>=0)  stcanon3++;
#ifndef FASTPROG
	if (r>=0 && ch>1-(CURDLEV<=printlev))  EMATDEBUGS(CURDLEV-1,ee,"\t\t\'\'\t");
	else if (r>=0 && ch>=0) {
		DEBUG(CURDLEV-1,">> extension of [%s] trying more: ",ESNAME(qn));
		for (i=0; i<ROWSM(ee); i++)  SDEBUG(CURDLEV-1," %s,",pfield_pvalue(6,EXPM(ee,i,COLSM(ee)-1),SIGNM(ee,i,COLSM(ee)-1)));
		SDEBUG(CURDLEV-1,"\n"); }
#endif
	if (r>=0)  r = genstep_sequencecheck_lev(2,qn,ee,1);
	if (r>=0)  r = genstep_structurecheck_lev(2,qn,ee,1);
	if (r>=0)  ststruc2++;
#ifndef FASTPROG
	if (r>=0 && IFRANDDEBUGLESS(55))  if (elimchm_represented_last(qn,ee,1)<0)
		{PROGERROR("The matrix must already be in the pfield at levels <=1 !");}
#endif
	if (r>=0)  r = genstep_sequencecheck_lev(1,qn,ee,1);
	if (r>=0)  r = genstep_structurecheck_lev(1,qn,ee,1);
	if (r>=0)  ststruc1++;
	if (r>=0 && GCTRL_CANONLEV(ctrl)<=1)  r = genstep_canonicalcheck_lev(1,qn,ee,1);
	if (r>=0)  stcanon1++;
	
		/**
		 * These are the full tests at levels 0:
		 * (See that genstep_seqsigncheck() was called in gener_extensions() above,
		 * and the line-check has been already performed at the start of this function...)
		 * Although the full tests should not be based on the last added line in general,
		 * we need to examine the last line for 3-connectivity in the sequence-check.
		**/
	if (r>=0)  r = genstep_sequencecheck(qn,ee,1);
	if (r>=0)  stseq0++;
	if (r>=0)  r = genstep_basischeck(qn,ee);
	if (r>=0)  r = genstep_structurecheck(qn);
	if (r>=0)  ststruc0++;
	if (r>=0 && GCTRL_CANONLEV(ctrl)==0)  r = genstep_canonicalcheck(qn);
	if (r>=0)  stcanon0++;
	
	ESMATRIX(qn) = eqo;
	return r>=0? 1:-1;
}

















/******************	Additions to generating	*******************/
/******************************************************************/


/**
 * The file gener-more.inc is provided for third-party additions to the generating process.
 * (See master description of generating in ../include/gener.h.)
 * The file (possibly) adds more fields to the elimination-sequence structure elimseqm
 * when included from eseq.h with GENER_ELIMQDEF>0.
 * When included from here with GENER_ELIMQDEF==0, the file
 * adds processing for more frame-options that control the generating process,
 * and adds more tests for the sequence and structure checks (see genstep.c).
 * 
**/

#define	GENER_ELIMQDEF 0
#include "gener-more.inc"
#undef	GENER_ELIMQDEF





































