
/********************************************************************************
 *										*
 *				 M A C E K					*
 *		("Matroids Also Computed Efficiently" Kit)			*
 *										*
 * A set of tools and routines for computations with representable matroids.	*
 * Copyright (C) 2001,2002  Petr Hlineny.					*
 *										*
 *   This program is free software; you can redistribute it and/or modify	*
 *   it under the terms of the GNU General Public License as published by	*
 *   the Free Software Foundation; either version 2 of the License, or		*
 *   (at your option) any later version.					*
 *   You should have received a copy of the GNU General Public License		*
 *   along with this program; if not, write to the Free Software		*
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.	*
 * 										*
 * See the top-level "ABOUT" and "doc/ *" files for more information.		*
 *										*
 ********************************************************************************/














/**
 * Some theory about bases and displayed minors for start...
 * Read general description of "structural" functions on the top of ../include/struct.h.
 * 
 * Since a matroid representation may display all different bases of the matroid,
 * we need a function that generates all the bases for the given matrix.
 * That means, for all nonzero subdeterminants of the matrix, we pivot the subdeterminant
 * and put the resulting matrix into the output list (including the original matrix).
 * We also provide "random" sublist of bases - a random selection from all bases for use
 * in randomized tests, that are requested by the parameter ch==-7 in some tests.
 * So far, we do not look at bases with isomorphic representations (- these may later be
 * factored out, but how can it be done?).
 * We allow "caching" of basis-list till the next call for all bases (only not randomized).
 * 
 * 
 * The word "minor" is used in the matroidal sense, but that has little meaning in
 * matrices alone.
 * Thus we speak about a "displayed minor" which corresponds to a submatrix in
 * a suitable representation obtained by a choice of the basis (via pivoting and scaling).
 * When reporting displayed minors, we factor out all minor automorphisms (it is faster),
 * so in order to get really all displayed-minor mappings, one must to apply
 * the minor automorphisms later.
 * 
 * In order to find whether a matroid has a given minor, we must search over all bases
 * of the matroid representation for a displayed minor similar to the searched minor.
 * As there are many bases, the search may be quite slow...
 * Moreover, we either have to make sure that the searched minor appears (unlabeled-)
 * equivalent the given minor representation - like if the representation is unique,
 * or to give a list of all inequivalent representations of the minor in the search.
 * 
 * .............................
 * 
 * 		(copy of this is also inside struct.h:)
 * 
 * *********** another way of looking at minors - try all deletions/contractions
 * 	- good when the size difference is small,
 * 	  but it needs a very fast equivalence testing
 * 
 * *********** implement hash-codes based on matroidal properties
 * 	(small-rank flats for fast, count bases per elements/pairs, etc for slower)
 * 	- use different levels of hashes - matroidal, fixed basis, fixed rows...
 * 
**/




#include "macek.h"
#include "str.h"











/******************	Finding bases in representations	*******************/
/**********************************************************************************/


//#undef CURDLEV
#define CURDLEV		7		/*8 (need <=DEBUGLEV+2 to perform paranoic tests) */



/**
 * The following function generates all displayed bases for the matroid given by e.
 * (This means that all possible multiple pivotings are applied to copies e.)
 * The new matrices equivalent to e are returned in a NULL terminated list of matrix copies,
 * in the same transposition state as given e.
 * The list starts with a copy of e itself, trimmed to the actual size.
 * If two distinct bases happen to produce the same matrix, it occurs twice.
 * The newly constructed equivalent matrices are NOT unit-scaled, they are just arbitrary.
 * (Remember to free the list and the matrices later.)
 * 
 * If ch>0, some more information are printed, and extra paranoic (debug) tests performed.
 * If ch==-7, then only a random sublist of bases is returned - for quick randomized pre-checks.
 * The list returned here is then supposed to be fed one by one to estruct_isminor...()
 * or in other functions testing for minors or equivalence.
 * 
 * How may one deal with isomorphic bases when looking for displayed minors ???
 * The best would be to find a way how to "guess" correct basis for the searched minor...
**/

#ifndef	STR_MAXBASES
#define	STR_MAXBASES	200000
#endif

ematrix**	struct_allbases_ext(int ch, ematrix *e) {
	ematrix 	*re, *ee, **bout;
	int		l;
	
	if (!e)  return NULL;
	if (ch>-7) {		/* allows to return previously cached list for non-randomized */
		bout = struct_allbases_reuse(e,-1,NULL);
		if (bout!=NULL)  return bout;
	}
	ee = ematrix_copy_to(e,ROWSM(e),COLSM(e));
	bout = alist_append(NULL,ematrix_copy(ee));	/* (the matrix itself is not recorded later) */
	DEBUG(CURDLEV,"Computing %s bases of the matrix %p ...\n",ch<=-7?"randomized":"all",e);
	if (ch>0)  EMATDEBUG(CURDLEV+2,e,"\t=\t");
	
	re = ematrix_refer_empty(ee);
	bout = struct_allbases_recur(ch,ee,re,0,0,bout);
	l = alist_getlength(bout);
	
	DEBUG(CURDLEV-2," := done computing %s %d bases of the matrix %p.\n",ch<=-7?"randomized":"all",l,e);
	dispose_ematrix(re);
	dispose_ematrix(ee);
	if (ch>-7)		/* allows to cache and reuse the list (but the list is stored later, not now !!!) */
		struct_allbases_reuse(NULL,l,bout);
	return bout;
}


/**
 * This is the recursive part of computing all bases for the given matrix e (struct_allbases_ext).
 * Its structure is similar to ematrix_determinant_recur().
 * We cycle recursively through all square submatrices of e, using a refering matrix re.
 * For each such matrix re, we try to pivot whole of it -- exchange its rows with its
 * columns in the basis (successful iff the determinant is nonzero, which is implictly tested).
 * Each successful pivoting results in recording a copy of the matrix.
 * Global static variables are used (for simplicity) to record the output list of matrices.
**/

ematrix**	struct_allbases_recur(int ch, ematrix *e, ematrix *re, int dr, int dc, ematrix **bout) {
	int		i,j;
	ematrix		*ee, *rre;
	sign_t		gg;
	
	DEBUG(CURDLEV+3,"entering estruct_allbases_recur() with dr=%d, dc=%d\n",dr,dc);
	if (ROWSM(re)>0) {
			/**
			 * The refering matrix re gives us a (nonempty) square submatrix
			 * of e, that is "inverted" here to display a new basis consisting
			 * of the rows not in re and columns in re.
			 * This works iff det(re)!=0 which is implicitly tested during pivoting.
			 * In fact, we pivot on a copy ee of e, and use a refering copy rre.
			 * The successfully pivoted lines are removed from rre, until whole
			 * rre is finished or there is a zero line (fails).
			**/
		ee = ematrix_copy(e);
		rre = ematrix_refer_all(re);  ematrix_rerefer(rre,ee);
#ifndef	FASTPROG
		if (ISREFMAT(e))  {PROGERROR("Do not call for a refering matrix e=%p.",e);}
		if (!ISREFMAT(re) || REFMAT(re)!=e)  {PROGERROR("Call for a refering matrix re -> e.");}
		if (ch>=0 && IFRANDDEBUG(88))
			if (!ematrix_isequal(re,rre)) {
				PROGERROR("these refering matrices must be equal!!!");
				//ematrix_print(re); ematrix_print(rre); ematrix_print(REFMAT(re)); ematrix_print(REFMAT(rre));
			}
		EMATDEBUG(CURDLEV+3,rre,"\t\t\t");
#endif
					/* pivots the chosen submatrix rre of the copy ee of e */
		while (COLSM(rre)>0) {
			for (i=ROWSM(rre)-1; i>=0; i--)
				if (SIGNM(rre,i,COLSM(rre)-1)!=0)  break;
			if (i<0)  break;
			ematrix_pivot(ee,GETREFMROW(rre,i),GETREFMCOL(rre,COLSM(rre)-1));
			ematrix_remove(rre,i,COLSM(rre)-1);
		}
		j = COLSM(rre);
		dispose_ematrix(rre);
		if (j==0) {		/* if the pivoting was successful, records the matrix */
			bout = alist_append(bout,ee);
			DEBUG(CURDLEV+0," - new basis accepted, stored in %p\n",ee);
			if (ch>0)  EMATDEBUG(CURDLEV+1,ee,"\t+\t");
			if (IFRANDDEBUG(444)) if (alist_getlength(bout)>STR_MAXBASES)
				{PROGERROREXIT("too many output bases >%d to handle",alist_getlength(bout));}
		} else {		/* if the pivoting was unsuccessful, disposes */
			DEBUG(CURDLEV+2," - pivoting unsuccessful for the chosen submatrix  \tXXX\n");
			dispose_ematrix(ee);
		}
#ifndef	FASTPROG
		if (ch>=0 && IFRANDDEBUG(66)) {
			ematrix_determinant(re,NULL,&gg);
			if ((gg==0)!=(j>0)) {
				PROGERROR("Unsuccessful pivoting for a (non)zero submatrix gg=%d, j=%d (or converse), why???",gg,j);
				EMATDEBUG(0,re,"!!\t");}
		}
#endif
	}
			/**
			 * Then we recursively try all descendant subsets of lines of e,
			 * these subsets are stored as square submatrices in re (which is reused).
			 * The parameters dr,dc store the first lines we should consider to
			 * be chosen into re in the next call to estruct_allbases_recur().
			 * Moreover, we allow randomization when ch==-7; then only a random
			 * subcollection of all bases is generated here.
			**/
	for (i=dr; i<ROWSM(e); i++)  for (j=dc; j<COLSM(e); j++) {
		ematrix_refadd_rc(re,i,j);
		bout = struct_allbases_recur(ch,e,re,i+1,j+1,bout);
		ematrix_remove(re,ROWSM(re)-1,COLSM(re)-1);
		if (ch<=-7) {
			if (COLSM(e)>4)  j += random()%(2+COLSM(e)/4);
			if (ROWSM(e)>4 && j>=COLSM(e)-1)  i += random()%(2+ROWSM(e)/4);
		}
	}
	return bout;
}



/**
 * This function allows to "re-use" once computed list of all bases.
 * For this to work, one has first to call this function from estruct_allbases_ext() with the
 * generated list in bl and the length in ll (used later to check correctness).
 * Then the user should call estruct_allbases_recycle() (a variant of this function) to recycle
 * the basis list instead of freeing it.
 * After that, the next call to estruct_allbases_ext() for the same matrix will return the
 * recycled list (really the same list, not a copy of it, so do not alter the list!).
**/

#if CACHEDBASES>0

ematrix**	struct_allbases_reuse(ematrix *e, int ll, ematrix **bl) {
	ematrix		**b;
	int		lx;
	static ematrix	**slist, **llast, *last;	/* (used to store the generated and re-used list, and some checks) */
	static int	llength = -1, firstcall = 1;
	
	if (firstcall) {	/* initial clearing of the cache */
		llength = -1;  firstcall = 0;
		last = NULL;  slist = llast = NULL;
	}
	if (ll>0) {		/* remembers the generated list for later check when recycling */
		if (last)  dispose_ematrix(last);
		last = ematrix_copy(bl[ll-1]);
		llength = ll;
		llast = bl;
	} else if (bl!=NULL) {	/* this part is called instead of freeing the basis list bl */
		for (b=bl; *b; b++) ;
				/* (must conform to the last remembered list from estruct_allbases_ext) */
		if (bl!=llast || llength!=b-bl || (last? !ematrix_isequal(last,b[-1]):1)) {
			if (bl)  dispose_alist_mats(bl);
			if (bl==llast)  {PROGERROR("damaged list when calling \"recycle\", shoud not happen");}
			else  DEBUG(0," :X not the last generated list when recycling bases\n");
		} else {
			if (slist)  dispose_alist_mats(slist);
			slist = bl;  llength = -1;
		}
	} else if (e!=NULL) {	/* and here the function returns a previously recycled list */
		if (slist)  if (ematrix_isequal(slist[0],e)) {
			lx = alist_getlength(slist);
			DEBUG(CURDLEV-2," := returning recycled %d bases of the matrix %p.\n",lx,e);
			struct_allbases_reuse(NULL,lx,slist);
			b = slist;  slist = NULL;
			return b;	/* (must clear the list reference before returning!) */
		}
	} else {
		firstcall = 1;	/* this NULL-call clears all cached lists from here */
		if (last)  dispose_ematrix(last);
		if (slist)  DEBUG(CURDLEV-1,"Freeing cached bases-list of length %d...\n",alist_getlength(slist));
		if (slist)  dispose_alist_mats(slist);
	}
	return NULL;		/* (the recycled list is possibly returned above) */
}

#else

ematrix**	struct_allbases_reuse(ematrix *e, int ll, ematrix **bl) {
	
	if (ll<0 && bl!=NULL)
		dispose_alist_mats(bl);
	return NULL;
}

#endif

















/******************	Finding displayed minors	*******************/
/**************************************************************************/


#undef CURDLEV
#define	CURDLEV		7		/*7 (need <=DEBUGLEV+2 to perform paranoic tests) */


/**
 * This function looks in the matrix e for displayed minors strongly-equivalent to emin.
 * In other words, it looks in all ordered subsets of rows and columns corresponding
 * to the dimensions of emin, whether the respective submatrix is a scale of emin.
 * Possible automorphic row- and column-permutations of emin are factored out,
 * so that one displayed minor comes out only once.
 * The function works only for connected minors(!), otherwise an error is issued.
 * The return list of matrices consists of refering (to e) matrices showing the emin-minor
 * exactly as it corresponds to the entries of emin, up to scaling of course.
 * The list is NULL terminated.
 * (Remember to free the list and the matrices later.)
 * 
 * Call this function only for a standalone (not refering) matrix, since otherwise the
 * returned list is not well-defined.
 * If emin is empty or bigger than e, then nothing is returned in a list.
 * If ch>0, some more information are printed, and extra paranoic tests are performed if ch>=0.
 * If ch==-3, then we only test for a minor, but do not generate any list (retun 0 or 1).
 * If ch==-2, then no minors are collected in the list, but all of them are still tested.
 * 
 * When looking for a displayed minor, we use quite complex hash-values for maps of rows
 * and of pairs of rows (minor -> matrix), which are described in estruct_dispminor_rowhash()
 * and estruct_dispminor_fitldet().
 * These values consider zero entries and zero 2x2 subdeterminants in the rows and the pairs.
 * The subdeterminants are computed also with respect to the "absolute values".
 * (Any mapping satisfying |a*b|=|a|*|b| works here - it preserves zero 2x2 determinants.)
 * In result, the hash values tell us what columns may be used in the displayed minor respecting
 * the selected rows (columns are considered unordered for now).
 * Once having a possible (ordered) row map from the minor to the matrix, we call the function
 * estruct_fitcols() to properly map the columns of the minor.
 * 
 * How may one deal with isomorphic bases when looking for displayed minors ???
 * The best would be to find a way how to "guess" correct basis for the searched minor...
 * 
**/

#ifndef	STR_MAXMINORS
#define	STR_MAXMINORS	30000
#endif
#define	MXMN	(NUM_LONG_BITS/2-2)		/* (must fit in long bits!) */
int	callhash, rejhash, rejrowf, rejcolf;	/* (for debugging statistics, here and in struct.c) */

FILE	*struct_printdispm = NULL;		/* request to print the displayed minors to the file */
char	*struct_printdispp = NULL;		/*  prefix for the print request */
int	struct_printdispa = 0;			/*  request to print all the displayed minors */
int	struct_printdispix = 0;			/*  minor-list index for the print request */


ematrix**	struct_dispminor_ext(int ch, ematrix *e, ematrix *emin) {
	ematrix		**mout;
	
	if (ch<-3)  {PROGERROR("Do not call with ch==%d <-3",ch);}
	if (ROWSM(e)>=MXMN || COLSM(e)>=MXMN)	/* (not practical for larger matrices anyway) */
		{PROGERROREXIT("not designed to deal with %d,%d > %d rows or columns",ROWSM(e),COLSM(e),MXMN);}
	if (ISREFMAT(e))  {PROGERROR("Do not call for a refering matrix e=%p.",e);}
	DEBUG(CURDLEV-1,"looking for %s displayed %p-minors (%dx%d) in %p (%dx%d)...\n",
			ch>-3?"all":"exi",emin,ROWSM(emin),COLSM(emin),e,ROWSM(e),COLSM(e));
	if (ch>0) { EMATDEBUG(CURDLEV+2,e,"\t");  EMATDEBUG(CURDLEV+2,emin,"\t>\t"); }
	mout = NULL;
	
	if (ROWSM(emin)<=ROWSM(e) && COLSM(emin)<=COLSM(e) && COLSM(emin)>0 && ROWSM(emin)>0) {
		mout = struct_dispminor_core(ch,e,emin,mout);	/* here is the real computation */
	}
	if (ch>-2) {
		DEBUG(CURDLEV-1-(mout!=NULL)," ... done, found %d displayed %p-minors in the matrix %p.\n",alist_getlength(mout),emin,e);
		if (mout && mout==(void*)1)  {PROGERROR("Should have collected the minors! %p",mout);}
	} else {
		DEBUG(CURDLEV-1-(mout!=NULL)," ... done, asked only for an existence of a minor, found %s.\n",mout!=NULL?"YES":"NO");
		if (mout && mout!=(void*)1)  {PROGERROR("Should have only indicated the existence of a minor! %p",mout);}
	}
	return mout;
}

			/* here is the real computation for struct_dispminor() ... */
ematrix**	struct_dispminor_core(int ch, ematrix *e, ematrix *emin, ematrix **mout) {
	int		i,j,k, br, hrr=0, pr[MXMN];
	long		hse[MXMN],hsmin[MXMN], hce[MXMN],hcmin[MXMN];
	long		*subca[MXMN+1],**subc, dsra[MXMN+1],*dsr;
	emxaut		**aut, **auti, **a;
	IHSETVARACCESS(emin,e);	/* supplementary variables for two-row mapping hash, see str.h */
	
	aut = auti = NULL;	/* (obtained only later on demand...) */
	callhash++;
		/**
		 * We first compute hash-values for pairs of rows of emin against e:
		 * Each entry IHGET(rm1,rm2,rb1,rb2) is an array assigned to the (possible)
		 * mapping of rows rm1,rm2 of emin (ordered!) to rows rb1,rb2 of e.
		 * The j-th entry of this array keeps a bitfield showing columns of e
		 * that may become j-th images of some columns of emin (no respect to their
		 * order in emin).
		 * Additionally, single-row hash codes are derived from the above values
		 * and stored in hsmin[],hse[] -- hsmin[i]==hse[j] if we can map row i->j.
		**/
	IHGETPOINTER = MMALLOC(IHGETSIZE*sizeof(IHGETPOINTER[0]));
	br = struct_dispminor_rowhash(ch,emin,e,hsmin,hse,IHGETPOINTER);
	if (br<0) {				/* if there is no row mapping possible */
		FREE(IHGETPOINTER);  if (br<0)  rejhash++;
		DEBUG(CURDLEV,"- row-pair hash test does not allow a minor occurence\n");
		return mout;
	}
	subca[0] = MMALLOC((ROWSM(emin)+1)*COLSM(emin)*sizeof(subca[0][0]));
	for (i=0; i<ROWSM(emin)+1; i++)  subca[i] = subca[0]+i*COLSM(emin);
	subc = subca+1;
	for (j=0; j<COLSM(emin); j++)  subc[-1][j] = (1l<<COLSM(e))-1;
	dsr = dsra+1;  dsr[-1] = 0;
	
		/**
		 * We now try all ordered mappings of rows of emin to rows of e that respect
		 * the above introduced hash-values for rows and their pairs.
		 * We use a standard backtracking technique to cycle through all mappings.
		 * We also check against the permutations of emin rows, see below.
		 * (The array subc[k][] keeps "accumulated" row-pair hash codes by the columns.)
		**/
	pr[0] = -1;  k = 0;
	while (k>=0) {
		if (k>=ROWSM(emin))  {PROGERROR("cannot get here!!! %d",k);}
		if (++pr[k]>=ROWSM(e)) {
			--k;  continue;			/* no more choices on this level */
		}
		if ((dsr[k-1]&(1l<<pr[k]))!=0)  continue;	/* (already mapped to pr[k]) */
		dsr[k] = dsr[k-1]|(1l<<pr[k]);
		DEBUG(CURDLEV+3,"%*smap considered pr[%d]=%d\n",2*k,"",k,pr[k]);
		if (hsmin[k]!=hse[pr[k]])  continue;	/* failed one-row hash */
		
		for (j=0; j<COLSM(emin); j++)  subc[k][j] = subc[k-1][j];
		for (j=0; j<COLSM(emin); j++) {
			for (i=0; i<k; i++)		/* check two row hash for all added pairs */
				subc[k][j] &= IHGET(i,k,pr[i],pr[k])[j];
			if (subc[k][j]==0)  break;
		}
		if (j<COLSM(emin))  continue;		/* when failed two-row hash */
#ifndef FASTPROG
		DEBUG(CURDLEV+2,"%*s  map found pr[%d]=%d%s\n",2*k,"",k,pr[k],k<ROWSM(emin)-1?"":"   \t--");
		if (CURDLEV+4<=DEBUGLEV) {
			SDEBUG(CURDLEV+4,"\t\t\t\t\t\t subc[] =");
			for (j=0; j<COLSM(emin); j++)  SDEBUG(CURDLEV+4," %lX,",subc[k][j]);
			SDEBUG(CURDLEV+4,"\n"); }
#endif
		if (k<ROWSM(emin)-1) {
			pr[++k] = -1;  continue;	/* continuing the search up */
		}
			/**
			 * Since we output the dispminors with their automorphisms factored out,
			 * we need to try one row mapping up to row permutations only.
			 * This is achieved by allowing only the unique "minimal" row selection
			 * from all permutations of the considered row map.
			 * (We later also factor out all emin-automorphisms fixing rows from
			 * selected columns.)
			**/
		if (aut==NULL) {		/* getting automorphisms of the minor emin */
			aut = struct_automorphisms_ch(ch<-1?-1:DECRCH(ch),emin,1);
			auti = NULL;		/* those automorphisms identical on the rows */
			for (a=aut; *a; a++) {
				for (j=br=0; j<ROWSM(emin); j++)  br |= ((*a)->pr[j]!=j);
				if (br==0)  auti = alist_append(auti,*a);
			}
			if (!auti)  {PROGERROR("Where is the identical automorphisms???");}
			DEBUG(CURDLEV+1,"  - using %d total and %d row-preserving automorphisms\n",alist_getlength(aut),alist_getlength(auti));
		}
		br = 0;				/* check whether an automorphism gives a "smaller" mapping */
		for (a=aut; *a && !br; a++)
			for (i=0; i<ROWSM(emin); i++) {
				if (pr[(*a)->pr[i]]>pr[i])  break;
				if (pr[(*a)->pr[i]]<pr[i]) { br = 1;  break; }
			}
		if (br)  DEBUG(CURDLEV+2,"X map not minimal with respect to row automorphisms\n");
		if (br)  continue;
		
			/**
			 * If all these tests pass well, then we call estruct_dispminor_rowfix()
			 * to look for the emin minor (by pr[]) in columns of e by brute force.
			 * The column hash-values hce,hcmin are computed there, not here for now.
			 * The row-fixing column permutations auti[] are used there.
			**/
#ifndef FASTPROG
		DEBUG(CURDLEV+1,"passed row selection, pr[] =");
		for (i=0; i<ROWSM(emin); i++)  SDEBUG(CURDLEV+1," %d,",pr[i]);
		SDEBUG(CURDLEV+1,"\n");
#endif
		/* the column hash values hcmin[],hce[] are not yet used here - set in rowfix() */
		mout = struct_dispminor_rowfix(ch,emin,e,hcmin,hce,pr,subc[k],auti,mout);
		hrr = 1;	/* (for debugging statistics only) */
		if (ch<=-3 && mout)  break;	/* (no output list is requested, just existence) */
	}
	
	if (!hrr)  rejrowf++;  else if (!mout)  rejcolf++;
	FREE(IHGETPOINTER);			/* frees the row hash memory allocated above */
	FREE(subca[0]);
	if (auti)  alist_free(auti);		/* frees all the automorphisms */
	if (aut)  struct_automorphisms_recycle(emin,aut);
	return mout;		/* (returns to struct_dispminor_ext()...) */
}







/******************	Minor-hash computation	*******************/
/******************************************************************/


/**
 * Here we compute hash values that could help to decide whether an (ordered) pair of rows
 * of the matrix em can be mapped to a pair of rows of eb within an occurence of a displayed
 * em-minor in eb; the values are stored in hspt (which is indexed by the macro IHGET()).
 * The entries of output hspt are indexed by (ordered) pairs of rows of em times (ordered) pairs of
 * rows of eb, and each entry format is described at estruct_dispminor_fitldet() below.
 * (Briefly speaking, the entries show to which columns of eb on these rows the columns of em
 * may be mapped - a bitfield for each j-th occurence of em in the columns of eb.)
 * The return value is normally 0, but it is -1 if no mapping of rows is possible.
 * Be careful that the matrices obtained here by applying "endomorphisms" may not be represented!
 * For use only in estruct_dispminor_ext().
**/

static int	maxabsvalix = 0;	/* the number of "abs values" (endomorphisms) we use here */

int	struct_dispminor_rowhash(int ch, ematrix *em, ematrix *eb, long hsm[], long hsb[], void *hspt) {
	
	ematrix		*ema[1+STR_MAXABSVALIX], *eba[1+STR_MAXABSVALIX];
	int		i,ii,j,jj, k,l, id,jd, ad,ax,ld, ret=0;
	int		rmat[MXMN][MXMN], rmin[MXMN][MXMN], ro[MXMN];
	int		*zprm[MXMN][MXMN][1+STR_MAXABSVALIX], *zpb[1+STR_MAXABSVALIX];
	long		zm[MXMN],zb[MXMN];
	emxaut		**aut,**a;	/* (just for debug testing...) */
	IHSETVARACCESS(em,eb);	/* supplementary variables for accessing two-row mapping hash */
	IHGETPOINTER = hspt;
	
	maxabsvalix = pfield_endomorph_number();
	if (maxabsvalix>STR_MAXABSVALIX)  maxabsvalix = STR_MAXABSVALIX;
	DEBUG(CURDLEV+1,"computing row-hash for %p and minor %p (using %d absv) ...\n",eb,em,maxabsvalix);
		/**
		 * We first identify zero entries and zero 2x2 subdeterminants of rows and
		 * pairs of rows in em,eb and in the endomorphism-valued copies ema,eba.
		 * (There may be more than one way to compute "endomorphism" - see maxabsvalix.)
		 * Be careful that the matrices obtained here by applying "endomorphisms"
		 * may not be properly represented!
		 * However, it is still correct to look at those invalid determinants as nonzeros.
		 * 
		 * We then use these zero-patterns to compute the IHGET (=*hspt) hash codes
		 * for pairs of rows of em times eb (see estruct_dispminor_fitldet).
		 * We also determine rmat[][] values showing which pairs of rows of eb
		 * may be images of pairs of rows of em.
		 * After that, we turn the hash-pairs into simplified one-line hash codes.
		**/
	ema[0] = em;  eba[0] = eb;
	for (l=0; l<maxabsvalix; l++) {
		ema[1+l] = ematrix_copy(em);  eba[1+l] = ematrix_copy(eb);
		for (i=0; i<ROWSM(em); i++)  for (j=0; j<COLSM(em); j++)
			pfield_endomorph(l,&EXPM(ema[1+l],i,j),&SIGNM(ema[1+l],i,j));
		for (i=0; i<ROWSM(eb); i++)  for (j=0; j<COLSM(eb); j++)
			pfield_endomorph(l,&EXPM(eba[1+l],i,j),&SIGNM(eba[1+l],i,j));
	}
			/* first compute zero-pattern for each row of em and of eb to bitfields */
	for (i=0; i<ROWSM(em); i++)
		for (j=0, zm[i]=0; j<COLSM(em); j++)
			if (SIGNM(em,i,j)!=0)  zm[i] |= (1l<<j);
	for (i=0; i<ROWSM(eb); i++)
		for (j=0, zb[i]=0; j<COLSM(eb); j++)
			if (SIGNM(eb,i,j)!=0)  zb[i] |= (1l<<j);
	
			/* then compute the pattern of zero 2x2-determinants for pairs of rows */
	zpb[0] = MMALLOC((1+maxabsvalix)*COLSM(eb)*sizeof(zpb[0][0]));
	for (l=0; l<1+maxabsvalix; l++)
		zpb[l] = zpb[0]+l*COLSM(eb);
	zprm[0][0][0] = MMALLOC(ROWSM(em)*ROWSM(em)*(1+maxabsvalix)*COLSM(em)*sizeof(zprm[0][0][0][0]));
	for (i=0; i<ROWSM(em); i++)  for (ii=i+1; ii<ROWSM(em); ii++) {
		for (l=0; l<1+maxabsvalix; l++)
			zprm[i][ii][l] = zprm[ii][i][l] =	/* (always symmetric on i,ii) */
				zprm[0][0][0]+((i*ROWSM(em)+ii)*(1+maxabsvalix)+l)*COLSM(em);
	}
	DEBUG(CURDLEV+4,"computing linedet for minor:\n");
	for (i=0; i<ROWSM(em); i++)  for (ii=0; ii<i; ii++) {
		for (l=0; l<1+maxabsvalix; l++)
			struct_dispminor_linedet(ema[l],0,i,ii,zprm[i][ii][l]);
	}
			/* finally, record what choices in eb agree by the zero patterns with em */
	DEBUG(CURDLEV+4,"computing linedet for major and pair-maps (min->maj):\n");
	for (i=0; i<ROWSM(em); i++)  for (ii=0; ii<ROWSM(em); ii++)
		rmin[i][ii] = 0;
	for (j=0; j<ROWSM(eb); j++)  for (jj=0; jj<j; jj++) {
		rmat[jj][j] = rmat[j][jj] = 0;
		for (l=0; l<1+maxabsvalix; l++)
			struct_dispminor_linedet(eba[l],0,j,jj,zpb[l]);
		for (i=0; i<ROWSM(em); i++)  for (ii=0; ii<ROWSM(em); ii++)  if (i!=ii) {
					/* (this cycle must use all pairs i<ii and i>ii) */
			DEBUG(CURDLEV+4,"@ at maj %d,%d min %d,%d(+rev) :\n",j,jj,i,ii);
			struct_dispminor_fitldet(COLSM(em),zm[i],zm[ii],zprm[i][ii],
						COLSM(eb),zb[j],zb[jj],zpb, IHGET(i,ii,j,jj));
			for (k=0; k<COLSM(em); k++)	/* (mapping is the same for both pairs reversed) */
				IHGET(ii,i,jj,j)[k] = IHGET(i,ii,j,jj)[k];
					/* marks when the pair j,jj can be mapped to: */
			if (IHGET(i,ii,j,jj)[0]!=0)  rmat[jj][j] = rmat[j][jj] = 1;
					/* marks when the pair i,ii can be mapped from: */
			if (IHGET(i,ii,j,jj)[0]!=0)  rmin[ii][i] = rmin[i][ii] = 1;
		}
	}
#ifndef FASTPROG
	DEBUG(CURDLEV+1,"... done computing row-hash for matrix %p and minor %p\n",eb,em);
	if (ch>0 && CURDLEV+2<=DEBUGLEV) {
		SDEBUG(CURDLEV,"\t\t\ttwo-row hash of em over all pairs mapped to:");
		for (j=0; j<ROWSM(em); j++)  for (jj=0; jj<ROWSM(em); jj++)
			SDEBUG(CURDLEV,"%s%c",jj==0?"\n\t\t\t\t":"  ",j==jj?' ':(rmin[j][jj]?'*':'.'));
		SDEBUG(CURDLEV,"\n\t\t\tend table\n");
		SDEBUG(CURDLEV,"\t\t\ttwo-row hash of eb over all pairs of the minor:");
		for (j=0; j<ROWSM(eb); j++)  for (jj=0; jj<ROWSM(eb); jj++)
			SDEBUG(CURDLEV,"%s%c",jj==0?"\n\t\t\t\t":"  ",j==jj?' ':(rmat[j][jj]?'*':'.'));
		SDEBUG(CURDLEV,"\n\t\t\tend table\n");
	}
	if (ch>=0 && IFRANDDEBUG(1111)) {
		aut = struct_automorphisms(em);		/* (aut-invariant test of IHGET) */
		for (a=aut,ax=0; *a; a++)  ax++;
		if (CURDLEV>DEBUGLEV) { id = 1;  jd = 1+(random()%ROWSM(eb))/3; }
		else { id = jd = 1; }
		for (j=0; j<ROWSM(eb); j+=jd)  for (jj=j+1; jj<ROWSM(eb); jj++)
		  for (i=0; i<ROWSM(em); i+=id)  for (ii=i+1; ii<ROWSM(em); ii++) {
			if (CURDLEV>DEBUGLEV) { ad = 1+random()%(ax+1);  ld = 1+random()%(COLSM(em)+1); }
			else { ad = ld = 1; }
			for (a=aut+1; a-aut<ax; a+=ad)  for (l=0; l<COLSM(em); l+=ld)
				if (IHGET((*a)->pr[i],(*a)->pr[ii],j,jj)[l]!=IHGET(i,ii,j,jj)[l])
					{PROGERROR("two-row hash is not automorphism-invariant:  j=%d, jj=%d,  i=%d->%d, ii=%d->%d,  l=%d;  IH %lX!=%lX",
						j,jj,i,(*a)->pr[i],ii,(*a)->pr[ii],l,IHGET(i,ii,j,jj)[l],IHGET((*a)->pr[i],(*a)->pr[ii],j,jj)[l]);}
		}
		struct_automorphisms_recycle(em,aut);
	}
#endif
		/**
		 * Here we look at all pairs of rows of em that may be mapped to some pairs
		 * of rows of eb - these form a graph described by rmin[][] wich must be complete.
		 * Then we look at all pairs of rows of eb that may be images of some pairs
		 * of rows of em - these form a graph described by rmat[][].
		 * An occurence of em-minor in eb must induce a clique in rmat, so we look
		 * whether rmat contains cliques of size >=ROWSM(em) by brute force.
		**/
	ret = 0;  l = 1;
	for (i=ii=0; i<ROWSM(em) && l; i++)  for (ii=0; ii<i && l; ii++)
		if (rmin[i][ii]==0)  l = 0;
	if (l==0) {
		DEBUG(CURDLEV+1,"no minor possible since the \"from\" two-row hash is not complete (%d,%d)\n",i-1,ii-1);
		ret = -1;
	}
	k = l = 0;  ro[0] = -1;
	while (k>=0 && ret>=0) {	/* (checking all ROWSM(em)-tuples of rows of eb for a clique) */
		if (++ro[k]>=ROWSM(eb)-ROWSM(em)+1+k) { --k;  continue; }
		for (j=0; j<k; j++)  if (rmat[ro[k]][ro[j]]==0)  break;
		if (j<k)  continue;
		if (k<ROWSM(em)-1) { ++k;  ro[k] = ro[k-1]; }
		else { l = 1;  break; }
	}
	if (l==0 && ret>=0) {
		DEBUG(CURDLEV+1,"no minor possible since no large clique in the \"to\" two-row hash\n");
		ret = -1;
	}
			/* we try to compute one-row hash ............ */
	for (i=0; i<ROWSM(em); i++)  hsm[i] = 0;
	for (j=0; j<ROWSM(eb); j++)  hsb[j] = 0;
	//********** what can be computed here ? - so that hsm[i]==hsb[j] if i->j in some mapping
	//	- the hash should be computed into a matrix hsm X hsb, not just arrays...
	
	ld = ax = ad = jd = id = ch = 0;
	FREE(zprm[0][0][0]);	/* disposing local computation data */
	FREE(zpb[0]);
	for (l=0; l<maxabsvalix; l++) {
		dispose_ematrix(eba[1+l]);  dispose_ematrix(ema[1+l]);
	}			/* (do not dipose eba[0]==eb, ema[0]==em !!!) */
	return ret;
}


/**
 * Here we find the clusters of zero 2x2 subdeterminants of the two given rows r1,r2 of ee.
 * (The zero subdeterminants form cliques if viewed as edges of a graph!)
 * The same is computed for columns if tr==1.
 * The result is stored in ar[] -- ar[i] contains the index of the first column j<i for
 * which the subdeterminant of columns i,j is zero, or ar[i]=i if there is no zero.
 * The columns of both-0 entries are treated specially - they form their own cluster.
 * For use only in estruct_dispminor_rowhash().
**/

void	struct_dispminor_linedet(ematrix *ee, int tr, int r1, int r2, int ar[]) {
	int	i,j,p, i0;
	sign_t	gg;
	
	if (tr)  ematrix_transpose(ee);
	i0 = -1;	/* (i0 keeps the first both-0 column, ar[0] starts with 0 automatically in the cycle...) */
	for (i=0; i<COLSM(ee); i++) {
		if (SIGNM(ee,r1,i)==0 && SIGNM(ee,r2,i)==0) {
			ar[i] = i0 = (i0==-1? i:i0);
			continue;	/* (both-0 columns need special treatment!) */
		}
		for (j=0; j<i; j++)  if (ar[j]==j && ar[j]!=i0) {
			p = ematrix_determinant_2x2ch(0,ee,r1,r2,j,i,NULL,&gg);
			if (p>=0 && gg==0)  break;
		}	/* (the subdeterminant may not be defined after endomorphisms, but then like nonzero) */
		ar[i] = j;
	}
#ifndef FASTPROG
	if (IFRANDDEBUG(44))  for (i=0; i<COLSM(ee); i++)
		if (ar[i]<0 || ar[i]>i)  {PROGERROR("something wrong with linedet");}
	if (CURDLEV+4<=DEBUGLEV) {
		SDEBUG(CURDLEV,"\t\t\tlinedet at %d,%d: ar[] = ",r1,r2);
		for (i=0; i<COLSM(ee); i++)  SDEBUG(CURDLEV," %d,",ar[i]);
		SDEBUG(CURDLEV,"\n");
	}
#endif
	if (tr)  ematrix_transpose(ee);
}

/**
 * Here we examine the zero-entry and zero-subdeterminant (see above) patterns zm1,zm2,
 * zb1,zb2 and *zpm[],*zpb[] that were previously taken from two row-pairs of two matrices.
 * (The zero-subdeterminant patterns are given as fields *zpm[],*zpb[] indexed first by
 * the index (<1+maxabsvalix) used to compute "absolute values" of entries for determinants
 *  - this is provided to reveal more differences between the two rows.)
 * The patterns zm1,zm2,zb1,zb2 simply have bits 1 for nonzero entries, the meaning
 * of *zpm[],*zpb[] is the same as described in estruct_dispminor_linedet().
 * 
 * If there is a possibility of a mapping from all cm columns of the minor to a subset cl[]
 * of cm (out of all cb) columns of the big matrix, the we record in bitfield ih[j]
 * the positions of the j-th elements of the subsets cl[].
 * "Possibility" is tested for the numbers of zeros in the single rows (rows are ordered here!),
 * and for the numbers of zero 2x2 subdeterminants in the row pairs.
 * (Both these numbers are permutation- and scale-invariants for proper "absolute values".)
 * For use only in estruct_dispminor_rowhash().
**/

void	struct_dispminor_fitldet(int cm, long zm1, long zm2, int *zpm[],
				int cb, long zb1, long zb2, int *zpb[], long ih[]) {
	int	j,k,l, pb, cl[MXMN], mm[MXMN];
	int	nzm1,nzm2, nzb1[MXMN],nzb2[MXMN], nzpm[1+STR_MAXABSVALIX];
	
	if (cm>cb || cb>MXMN)  {PROGERROR("something wrong with the numbers of columns %d > %d > %d",cm,cb,MXMN);}
	if (zm1==0 || zm2==0 || zb1==0 || zb2==0)  {PROGERROR("something wrong - rows may not be all zeros %lX,%lX,%lX,%lX",zm1,zm2,zb1,zb2);}
	
			/* initialize the output hash, and count the numbers of zeros in the minor rows */
	for (nzm1=nzm2 = j=0; j<cm; j++) {
		ih[j] = 0l;
		if ((zm1&(1l<<j))==0)  nzm1++;
		if ((zm2&(1l<<j))==0)  nzm2++;
	}		/* also count the numbers of zero-subdeterminants in the minor rows by l */
	for (l=0; l<1+maxabsvalix; l++) {
		for (j=0; j<cm; j++)  mm[zpm[l][j]] = 0;
		nzpm[l] = 0;
		for (j=0; j<cm; j++)  nzpm[l] += mm[zpm[l][j]]++;
	}
	k = 0;  cl[0] = -1;
	while (k>=0) {
			/* here we generate all cm-elem subsets of cb columns for the hash computation */
		++cl[k];
		if (cl[k]>=cb-cm+1+k) { --k;  continue; }
			/* we compute the numbers of zeros in the selected columns along the way */
		nzb1[k] = (k>0? nzb1[k-1]:0) + ((zb1&(1l<<cl[k]))==0);
		nzb2[k] = (k>0? nzb2[k-1]:0) + ((zb2&(1l<<cl[k]))==0);
		if (nzb1[k]>nzm1 || nzb2[k]>nzm2)  continue;
			/* while not all columns are selected, we continue further */
		if (k<cm-1) {
			++k;  cl[k] = cl[k-1];  continue;
		}
			/* the final test whether the numbers of zeros agree for the selection */
		if (nzb1[k]!=nzm1 || nzb2[k]!=nzm2)  continue;
		
			/* now we compare numbers of zero subdeterminants (at index l=0..maxabsvalix) */
		for (l=0; l<1+maxabsvalix; l++) {
			for (j=0; j<cm; j++)  mm[ zpb[l][cl[j]] ] = 0;
			pb = 0;
			for (j=0; j<cm; j++)  pb += mm[ zpb[l][cl[j]] ]++;
			if (pb!=nzpm[l])  break;	/* (the numbers must be equal at all l's to pass) */
		}
		if (l<1+maxabsvalix)  continue;
		
		for (j=0; j<cm; j++)	/* if there was no contradiction above, we record the subset */
			ih[j] |= (1l<<cl[j]);
	}
#ifndef FASTPROG
	if (IFRANDDEBUG(140)) {
		SDEBUG(CURDLEV+4,"\t\t\t\tfitldet ih[] = ");
		if (CURDLEV+4<=DEBUGLEV)  for (j=0; j<cm; j++)  SDEBUG(CURDLEV+4," %lX,",ih[j]);
		SDEBUG(CURDLEV+4,"\n");
		for (j=0; j<cm; j++)  if (ih[j]<0 || ih[j]>=(1l<<cb))  {PROGERROR("Wrong output in fitldet");}
	}
#endif
}







/******************	Minors on fixed rows	*******************/
/******************************************************************/


/**
 * This function looks for all displayed em-minors in eb that map the rows of em by the given
 * mapping prm[], factoring out all fixed-row column automorphisms of em given in arow.
 * Additionally, cmpat[j] gives a bit pattern of columns of eb that may be j-th images of
 * columns of em in eb (bitwise-and over all pairs of rows, as computed above).
 * 
 * The arrays hcb[],hcm[] are for single-row mapping hash, but they are not set in advance,
 * so they are computed here from scratch (this may change later...).
 * If ch>0, then matrices are printed, and if ch<0, then no paranoic debug tests are executed.
 * If ch==-2, then no minors are collected, and if ch<=-3, then only the first minor is taken.
 * The function estruct_fitcols() from struc.c is used here, and so the minor must be
 * connected as required in that function.
 * For use only in estruct_dispminor_ext().
**/

ematrix**	struct_dispminor_rowfix(int ch, ematrix *em, ematrix *eb, long hcm[], long hcb[],
					int prm[], long cmpat[], emxaut **arow, ematrix **mout) {
	ematrix		*reb, *eo, *ech=NULL;
	emxaut		**a;
	int		i,j,jj, p,r,br, pcm[MXMN], ln,lold,ld;
	exp_t		xr[MXMN],xc[MXMN];
	sign_t		gr[MXMN],gc[MXMN];
	void		*data;
	char		buf[100];
	
	ld = lold = (ch>-2? alist_getlength(mout):0);
	if (lold>STR_MAXMINORS)  {PROGERROREXIT("too many output minors >=%d to handle",lold);}
		/**
		 * The one-column hash computed above just lists columns of eb that are
		 * possible images based on given pattern cmpat[], no more distinction is computed.
		 * So far, there is no two-column hash supplied to estruct_fitcols()...
		 * Remember that the one-column hash is refined by considering zero-entries
		 * of columns later in estruct_fitcols().
		**/
	for (j=0; j<COLSM(em); j++)  hcm[j] = 1;
	for (j=0; j<COLSM(eb); j++)  hcb[j] = 0;
			/* looking for columns of eb that can be subsets for em according to cmpat */
	for (j=0; j<COLSM(em); j++) {
		p = (j==0? -1:-2);
		for (jj=0; jj<COLSM(eb); jj++) {
			if (p>=-1 && (cmpat[j]&(1l<<jj))!=0) {
				hcb[jj] = 1;  p = jj;	/* marks possible column selections */
			}
			if (p<-1 && (cmpat[j-1]&(1l<<jj))!=0)
				p = -1;
		}
		if (p<0)  break;	/* if j-th column cannot be selected in proper place */
		for (jj=p+1; jj<COLSM(eb); jj++)  hcb[jj] = 0;
	}
	if (j<COLSM(em)) {		/* if some col index cannot be selected for an image of em */
		DEBUG(CURDLEV+2,"no column selection possible \tXXX\n");
		return mout;
	}
			/* constructing refering matrix that exacty matches the given row map in eb */
	reb = ematrix_refer(eb,-1,-1,0,COLSM(eb));
	for (i=0; i<ROWSM(em); i++)
		ematrix_refadd_row(reb,prm[i]);
#ifndef	FASTPROG
	if (CURDLEV+2<=DEBUGLEV) {
		DEBUG(CURDLEV+2,"eb col sel =");
		for (jj=0; jj<COLSM(eb); jj++)  SDEBUG(CURDLEV+2," %ld,",hcb[jj]);
		SDEBUG(CURDLEV+2," \tOK\n");
		for (j=0; j<COLSM(em); j++)  if (cmpat[j]==0)  {PROGERROR("should not get cmpat[j]==0");}
	}
	if (ch>0 && CURDLEV+4<=DEBUGLEV) {
		EMATDEBUGS(CURDLEV+4,reb,"\t\t\t@\t");  EMATDEBUGS(CURDLEV+4,eb,"\t\t\t->\t");
	}
#endif
		/**
		 * We have constructed a refering matrix reb from selected rows of eb, and then we
		 * call estruct_fitcols() to find all mappings of columns of em to columns of reb.
		 * The columns mappings are returned in pcm[], and the associated line scales
		 * are silently discarded (used only in debug tests).
		 * So far, there is no two-column hash supplied to estruct_fitcols()...
		**/
	r = struct_fitcols(1,em,reb,1,hcm,hcb,NULL,NULL, pcm,xr,gr,xc,gc,&data);
	if (r<0)  DEBUG(CURDLEV+2,"rejected by fitcols init   XXX\n");
	while (r>=0) {
		r = struct_fitcols(0,em,reb,1,hcm,hcb,NULL,NULL, pcm,xr,gr,xc,gc,&data);
		if (r<0)  break;		/* if there is no more col mapping */
		br = 0;
		for (a=arow; *a && !br; a++)	/* tests against automorphic column permutations of em */
			for (j=0; j<COLSM(em) && !br; j++) {
				if (pcm[(*a)->pc[j]]>pcm[j])  break;
				if (pcm[(*a)->pc[j]]<pcm[j])  br = 1;
			}
		if (br!=0)  continue;
						/* recording the minor as a refering submatrix to eb */
		eo = ematrix_refer_empty(eb);
		for (i=0; i<ROWSM(em); i++)		/* (rows of the minor) */
			ematrix_refadd_row(eo,prm[i]);
		for (j=0; j<COLSM(em); j++)		/* (columns of the minor) */
			ematrix_refadd_col(eo,pcm[j]);
#ifndef	FASTPROG
		DEBUG(CURDLEV+2,"recording #%d new minor %p\n",ld++,eo);
		if (ch>0)  EMATDEBUGS(CURDLEV+2,eo,"\t\t>\t");
		if (ch>=0 && IFRANDDEBUG(333)) {	/* paranoic tests before accepting */
			for (j=0; j<COLSM(em); j++)
				if (pcm[j]<0 || pcm[j]>=COLSM(eb))  {PROGERROR("not a column permutation for a minor, why???");}
			ech = ematrix_copy(eo);
			for (i=0; i<ROWSM(em); i++)  for (j=0; j<COLSM(em); j++) {
				ematrix_multiply_entry(ech,i,j,-1,xr[i],gr[i]);
				ematrix_multiply_entry(ech,i,j,-1,xc[pcm[j]],gc[pcm[j]]);
			}
			if (!ematrix_isequal(em,ech)) {
				PROGERROR("the matrix found is not the same as the minor!");
				EMATDEBUG(0,em,"\t");  EMATDEBUG(0,ech," !\t"); }
			dispose_ematrix(ech);
		}
#endif
		if (struct_printdispm) {	/* should print the minor to struct_printdispm: */
			fprintf(struct_printdispm,"%srecording #%d (at %d) new minor %p [%s]%s:\n",
				struct_printdispp,(struct_printdispa<=0?1:struct_printdispa++),struct_printdispix,eo,
				EMNAME(em),(ch<=-3?" (only the first one)":""));
			fprintf(struct_printdispm,"%s  rmul",struct_printdispp);
			for (i=0; i<ROWSM(em); i++)  fprintf(struct_printdispm,"\t  %s",pfield_pvalue(7,xr[i],gr[i]));
			fprintf(struct_printdispm,"\n%s  cmul",struct_printdispp);
			for (j=0; j<COLSM(em); j++)  fprintf(struct_printdispm,"\t  %s",pfield_pvalue(7,xc[pcm[j]],gc[pcm[j]]));
			fprintf(struct_printdispm,"\n");
			snprintf(buf,80,"%s\t",struct_printdispp);  buf[80]=0;
			ematrix_fprint_nofr(struct_printdispm,eo,buf);
		}
		if (ch<=-2) {			/* if no output list is requested, just existence */
			dispose_ematrix(eo);  mout = (ematrix**)1;
			if (ch<=-3)  break;
		} else {			/* adds to the output list of matrices otherwise */
			mout = alist_append(mout,eo);
		}
	}
			/* destroying locally-used data, also data used by estruct_fitcols() */
	struct_fitcols(-1,em,reb,1,hcm,hcb,NULL,NULL, pcm,xr,gr,xc,gc,&data);
	dispose_ematrix(reb);
	ln = (ch>-2? alist_getlength(mout): (mout!=NULL));
	if (ln-lold>0)  DEBUG(CURDLEV+1," - found %d of %p-minors for given row selection\n",ln-lold,em);
	else  DEBUG(CURDLEV+2," - found NO %p-minors for given row selection\n",em);
	return mout;
}


























/******************	Testing all (pivoted) minors	*******************/
/**************************************************************************/


#undef CURDLEV
#define	CURDLEV		7


/**
 * This function looks in the matrix e for all (pivoted) minors isomorphic to a member
 * of the given list lmin (in fact, only equivalent to the given representation of the minor!).
 * It examines all bases of e against all matrices in the list.
 * The return value is simple yes-no (1 or 0), no list is generated or returned.
 * A version of this function getting just one minor instead of a list is below.
 * If ch>0, some more information are printed.
 * If ch==-7, then only a randomized test is performed - for only a sublist of bases.
 * There are no extra checks for any value of ch.
**/

int	struct_hasminorlist_ext(int ch, ematrix *e, ematrix **lmin) {
	int		n, ret=0;
	ematrix		**lm, **bas,**b;
	//extern int	callhash, rejhash, rejrowf, rejcolf;	/* (for debugging statistics) */
	
	callhash = rejhash = rejrowf = rejcolf = 0;
	n = alist_getlength(lmin);
#ifndef	FASTPROG
	DEBUG(CURDLEV-2,"Going to test %p for %s (pivoted) minors on the %d-list...\n",e,ch<=-7?"randomized":"all",n);
	if (ch>0) {
		EMATDEBUG(CURDLEV+1,e,"\t\t");
		for (lm=lmin; *lm; lm++)  EMATDEBUG(CURDLEV+1,*lm,"\t\t>\t"); }
#endif
	if (n<=0)  return 0;
	//********* decide whether better to transpose the matrices...........
	
	ret = 0;	/* testing displayed minors over all possible bases (or randomized ch==-7) */
	if (ch>-7)  bas = struct_allbases_ch(DECRCH(ch),e);
	else  bas = struct_allbases_rand(e);	/* (for ch==-7, we test only a randomized list of bases - faster) */
	for (b=bas; *b && (ret<=0||struct_printdispa>0); b++) {
		DEBUG(CURDLEV+1,"# matrix base %p testing for minor %d-list\n",*b,n);
		if (ch>1)  EMATDEBUG(CURDLEV+2,*b,"\t\t~\t");
			/* testing minors from the list one by one */
		for (lm=lmin; *lm && (ret<=0||struct_printdispa>0); lm++) {
			struct_printdispix = (int)(lm-lmin)+1;
			if (struct_printdispa>0)  ret = struct_isdispminor_pr(*b,*lm) ?1:ret;
			else  ret = struct_isdispminor(*b,*lm) ?1:ret;
		}	/* (spec. only tests minors, no list is returned) */
	}
	if (ch>-7) {	/* when dealing with full list of bases, we may "recycle" it for further use */
		struct_allbases_recycle(bas);  bas = NULL;
	} else {
		dispose_alist_mats(bas);
	}
	DEBUG(CURDLEV-2,"... done, matrix %p testing for %s %d-list minors - %s (%d)\n",e,ch<=-7?"randomized":"all",n,ret?"YES":"NO",(int)(lm-lmin));
	DEBUG(CURDLEV-2,"\t(dispminor stat: hash %d, rej %d, rej rowf %d, rej colf %d)\n",callhash,rejhash,rejrowf,rejcolf);
	struct_printdispm = NULL;  struct_printdispp = NULL;  struct_printdispa = 0;
	return ret? 1:0;
}

int	struct_hasminor_ext(int ch, ematrix *e, ematrix *emin) {
	ematrix	**el;
	int	r;

	DEBUG(CURDLEV-1,"The minor %p is stored in a 1-list for testing.\n",emin);
	el = alist_append(NULL,emin);
	r = struct_hasminorlist_ext(ch,e,el);
	alist_free(el);
	return r;
}


/**
 * It is possible to request printing the minor found in the dispminor test above...
 * 
**/

void	struct_hasminor_printit_ext(FILE *fp, char *pref, int a) {
	
	struct_printdispm = fp;
	struct_printdispp = pref?pref:"";
	struct_printdispa = a;
}






































